#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

"""Discover and lookup command plugins."""

from collections.abc import Iterable, Iterator
import importlib.metadata
import logging
from typing import TypeAlias

import stevedore

from cliff import command

LOG = logging.getLogger(__name__)


def _get_commands_by_partial_name(
    args: list[str], commands: list[str]
) -> list[str]:
    n = len(args)
    candidates = []
    for command_name in commands:
        command_parts = command_name.split()
        if len(command_parts) != n:
            continue
        if all(command_parts[i].startswith(args[i]) for i in range(n)):
            candidates.append(command_name)
    return candidates


class EntryPointWrapper:
    """An entrypoint-like object.

    Wrap up a command class already imported to make it look like a plugin.
    """

    def __init__(
        self, name: str, command_class: type[command.Command]
    ) -> None:
        self.name = name
        self.command_class = command_class

    def load(self) -> type[command.Command]:
        return self.command_class

    @property
    def value(self) -> str:
        # fake entry point target for logging purposes
        return ':'.join(
            [self.command_class.__module__, self.command_class.__name__]
        )


EntryPointT: TypeAlias = EntryPointWrapper | importlib.metadata.EntryPoint


class CommandManager:
    """Discovers commands and handles lookup based on argv data.

    :param namespace: String containing the entrypoint namespace for the
        plugins to be loaded. For example, ``'cliff.formatter.list'``.
        If provided, commands will be loaded from this namespace initially,
        though it remains possible to load from additional namespaces with
        :meth:`CommandManager.load_commands`.
    :param convert_underscores: Whether cliff should convert underscores to
        spaces in entry_point commands.
    :param ignored_modules: A list of module names to ignore when loading
        commands. This will be matched partial, so be as specific as needed.
    """

    def __init__(
        self,
        namespace: str | None = None,
        convert_underscores: bool = True,
        *,
        ignored_modules: Iterable[str] | None = None,
    ) -> None:
        self.namespace = namespace
        self.convert_underscores = convert_underscores
        self.ignored_modules = ignored_modules or ()

        self.commands: dict[str, EntryPointT] = {}
        self._legacy: dict[str, str] = {}
        self.group_list: list[str] = []
        self._load_commands()

    def _load_commands(self) -> None:
        # NOTE(jamielennox): kept for compatibility.
        if self.namespace:
            self.load_commands(self.namespace)

    @staticmethod
    def _is_module_ignored(
        module_name: str, ignored_modules: Iterable[str]
    ) -> bool:
        # given module_name = 'foo.bar.baz:wow', we expect to match any of
        # the following ignores: foo.bar.baz:wow, foo.bar.baz, foo.bar, foo
        while True:
            if module_name in ignored_modules:
                return True

            split_index = max(module_name.rfind(':'), module_name.rfind('.'))
            if split_index == -1:
                break

            module_name = module_name[:split_index]

        return False

    def load_commands(self, namespace: str) -> None:
        """Load all the commands from an entrypoint

        :param namespace: The namespace to load commands from.
        :returns: None
        """
        self.group_list.append(namespace)
        em: stevedore.ExtensionManager[command.Command]
        # note that we don't invoke stevedore's conflict resolver functionality
        # because that is namespace specific and we care about conflicts
        # regardless of the namespace
        em = stevedore.ExtensionManager(namespace)
        for ext in em:
            LOG.debug('found command %r', ext.name)

            if self._is_module_ignored(ext.module_name, self.ignored_modules):
                LOG.debug(
                    'extension found in ignored module %r: skipping',
                    ext.module_name,
                )
                continue

            cmd_name = (
                ext.name.replace('_', ' ')
                if self.convert_underscores
                else ext.name
            )

            if cmd_name in self.commands:
                # Attention, programmers: If you arrived here attempting to
                # resolve a warning in your application then you have a command
                # with the same name either defined more than once in the same
                # application (a typo?) or in multiple packages (for example,
                # a package that adds plugins to your applications). The latter
                # can often happen if you e.g. move a command from a plugin to
                # the core application. In this situation, you should add the
                # old location to 'ignored_modules' and the plugin will now be
                # ignored.
                LOG.warning(
                    'found duplicate implementations of the %(name)r command '
                    'in the following modules: %(modules)s: this is likely '
                    'programmer error and should be reported as a bug to the '
                    'relevant project(s)',
                    {
                        'name': cmd_name,
                        'modules': ', '.join(
                            [
                                self.commands[cmd_name].value,
                                ext.entry_point.value,
                            ]
                        ),
                    },
                )

            self.commands[cmd_name] = ext.entry_point

    def __iter__(
        self,
    ) -> Iterator[tuple[str, EntryPointT]]:
        return iter(self.commands.items())

    def add_command(
        self, name: str, command_class: type[command.Command]
    ) -> None:
        self.commands[name] = EntryPointWrapper(name, command_class)

    def add_legacy_command(self, old_name: str, new_name: str) -> None:
        """Map an old command name to the new name.

        :param old_name: The old command name.
        :type old_name: str
        :param new_name: The new command name.
        :type new_name: str
        """
        self._legacy[old_name] = new_name

    def find_command(
        self, argv: list[str]
    ) -> tuple[type[command.Command], str, list[str]]:
        """Given an argument list, find a command and
        return the processor and any remaining arguments.
        """
        start = self._get_last_possible_command_index(argv)
        for i in range(start, 0, -1):
            name = ' '.join(argv[:i])
            search_args = argv[i:]
            # The legacy command handling may modify name, so remember
            # the value we actually found in argv so we can return it.
            return_name = name
            # Convert the legacy command name to its new name.
            if name in self._legacy:
                name = self._legacy[name]

            found = None
            if name in self.commands:
                found = name
            else:
                candidates = _get_commands_by_partial_name(
                    argv[:i], list(self.commands.keys())
                )
                if len(candidates) == 1:
                    found = candidates[0]
            if found:
                cmd_ep = self.commands[found]
                cmd_factory = cmd_ep.load()
                return (cmd_factory, return_name, search_args)
        else:
            raise ValueError(f'Unknown command {argv!r}')

    def _get_last_possible_command_index(self, argv: list[str]) -> int:
        """Returns the index after the last argument
        in argv that can be a command word
        """
        for i, arg in enumerate(argv):
            if arg.startswith('-'):
                return i
        return len(argv)

    def add_command_group(self, group: str | None = None) -> None:
        """Adds another group of command entrypoints"""
        if group:
            self.load_commands(group)

    def get_command_groups(self) -> list[str]:
        """Returns a list of the loaded command groups"""
        return self.group_list

    def get_command_names(self, group: str | None = None) -> list[str]:
        """Returns a list of commands loaded for the specified group"""
        group_list: list[str] = []
        if group is not None:
            em: stevedore.ExtensionManager[command.Command]
            em = stevedore.ExtensionManager(group)
            for ep in em:
                cmd_name = (
                    ep.name.replace('_', ' ')
                    if self.convert_underscores
                    else ep.name
                )
                group_list.append(cmd_name)
            return group_list
        return list(self.commands.keys())



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import argparse
from collections.abc import Sequence
import inspect
import traceback
from typing import Any

import autopage.argparse

from cliff import command


class HelpExit(SystemExit):
    """Special exception type to trigger quick exit from the application

    We subclass from SystemExit to preserve API compatibility for
    anything that used to catch SystemExit, but use a different class
    so that cliff's Application can tell the difference between
    something trying to hard-exit and help saying it's done.
    """


class HelpAction(argparse.Action):
    """Provide a custom action so the -h and --help options
    to the main app will print a list of the commands.

    The commands are determined by checking the CommandManager
    instance, passed in as the "default" value for the action.
    """

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ) -> None:
        app = self.default
        pager = autopage.argparse.help_pager(app.stdout)
        color = pager.to_terminal()
        autopage.argparse.use_color_for_parser(parser, color)
        with pager as out:
            parser.print_help(out)
            title_hl = ('\033[4m', '\033[0m') if color else ('', '')
            out.write('\n{}Commands{}:\n'.format(*title_hl))
            dists_by_module = command._get_distributions_by_modules()

            def dist_for_obj(obj: object) -> str | None:
                mod = inspect.getmodule(obj)
                if mod is None:
                    raise RuntimeError(f"failed to find app: {obj}")
                name = mod.__name__.partition('.')[0]
                return dists_by_module.get(name)

            app_dist = dist_for_obj(app)
            command_manager = app.command_manager
            for name, ep in sorted(command_manager):
                try:
                    factory = ep.load()
                except Exception:
                    out.write(f'Could not load {ep!r}\n')
                    if namespace.debug:
                        traceback.print_exc(file=out)
                    continue
                try:
                    kwargs = {}
                    fact_args = inspect.getfullargspec(factory.__init__).args
                    if 'cmd_name' in fact_args:
                        kwargs['cmd_name'] = name
                    cmd = factory(app, None, **kwargs)
                    if cmd.deprecated:
                        continue
                except Exception as err:
                    out.write(f'Could not instantiate {ep!r}: {err}\n')
                    if namespace.debug:
                        traceback.print_exc(file=out)
                    continue
                one_liner = cmd.get_description().split('\n')[0].rstrip('.')
                dist_name = dist_for_obj(factory)
                if dist_name and dist_name != app_dist:
                    dist_info = ' (' + dist_name + ')'
                    if color:
                        dist_info = f'\033[90m{dist_info}\033[39m'
                else:
                    dist_info = ''
                padded_name = f'{name:<13}'
                if color:
                    padded_name = f'\033[36m{padded_name}\033[39m'
                out.write(f'  {padded_name}  {one_liner}{dist_info}\n')
        raise HelpExit()


class HelpCommand(command.Command):
    """print detailed help for another command"""

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'cmd',
            nargs='*',
            help='name of the command',
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> int:
        if parsed_args.cmd:
            try:
                the_cmd = self.app.command_manager.find_command(
                    parsed_args.cmd,
                )
                cmd_factory, cmd_name, search_args = the_cmd
            except ValueError:
                # Did not find an exact match
                cmd = parsed_args.cmd[0]
                fuzzy_matches = [
                    k[0]
                    for k in self.app.command_manager
                    if k[0].startswith(cmd)
                ]
                if not fuzzy_matches:
                    raise
                self.app.stdout.write(f'Command "{cmd}" matches:\n')
                for fm in sorted(fuzzy_matches):
                    self.app.stdout.write(f'  {fm}\n')
                return 0
            kwargs = {}
            if 'cmd_name' in inspect.getfullargspec(cmd_factory.__init__).args:
                kwargs['cmd_name'] = cmd_name
            cmd = cmd_factory(self.app, self.app_args, **kwargs)
            full_name = (
                cmd_name
                if self.app.interactive_mode
                else ' '.join([self.app.NAME, cmd_name])
            )
            cmd_parser = cmd.get_parser(full_name)
            pager = autopage.argparse.help_pager(self.app.stdout)
            with pager as out:
                autopage.argparse.use_color_for_parser(
                    cmd_parser, pager.to_terminal()
                )
                cmd_parser.print_help(out)
        else:
            action = HelpAction([], dest=argparse.SUPPRESS, default=self.app)
            action(self.app.parser, self.app.options, None, None)
        return 0



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

"""Application base class."""

import itertools
import shlex
import sys
from typing import TYPE_CHECKING, TextIO

import autopage.argparse
import cmd2

if TYPE_CHECKING:
    from . import app as _app
    from . import commandmanager as _commandmanager


class InteractiveApp(cmd2.Cmd):
    """Provides "interactive mode" features.

    Refer to the cmd2_ and cmd_ documentation for details
    about subclassing and configuring this class.

    .. _cmd2: https://cmd2.readthedocs.io/en/latest/
    .. _cmd: http://docs.python.org/library/cmd.html

    :param parent_app: The calling application (expected to be derived
                       from :class:`cliff.app.App`).
    :param command_manager: A :class:`cliff.commandmanager.CommandManager`
                            instance.
    :param stdin: Standard input stream
    :param stdout: Standard output stream
    """

    use_rawinput = True
    doc_header = "Shell commands (type help <topic>):"
    app_cmd_header = "Application commands (type help <topic>):"

    def __init__(
        self,
        parent_app: '_app.App',
        command_manager: '_commandmanager.CommandManager',
        stdin: TextIO | None,
        stdout: TextIO | None,
        errexit: bool = False,
    ):
        self.parent_app = parent_app
        if not hasattr(sys.stdin, 'isatty') or sys.stdin.isatty():
            self.prompt = f'({parent_app.NAME}) '
        else:
            # batch/pipe mode
            self.prompt = ''
        self.command_manager = command_manager
        self.errexit = errexit
        cmd2.Cmd.__init__(self, 'tab', stdin=stdin, stdout=stdout)

    # def _split_line(self, line: cmd2.Statement) -> list[str]:
    def _split_line(self, line: str | cmd2.Statement) -> list[str]:
        # cmd2 >= 0.9.1 gives us a Statement not a PyParsing parse
        # result.
        parts = shlex.split(line)
        if isinstance(line, cmd2.Statement):
            parts.insert(0, line.command)
        return parts

    def default(self, line: str) -> bool | None:
        # Tie in the default command processor to
        # dispatch commands known to the command manager.
        # We send the message through our parent app,
        # since it already has the logic for executing
        # the subcommand.
        line_parts = self._split_line(line)
        ret = self.parent_app.run_subcommand(line_parts)
        if self.errexit:
            # Only provide this if errexit is enabled,
            # otherise keep old behaviour
            return bool(ret)
        return None

    def completedefault(  # type: ignore[override]
        self, text: str, line: str, begidx: int, endidx: int
    ) -> list[str]:
        """Default tab-completion for command prefix with completer delimiter.

        This method filters only cliff style commands matching provided
        command prefix (line) as cmd2 style commands cannot contain spaces.
        This method returns text + missing command part of matching commands.
        This method does not handle options in cmd2/cliff style commands, you
        must define complete_$method to handle them.
        """
        return [x[begidx:] for x in self._complete_prefix(line)]

    def _complete_prefix(self, prefix: str) -> list[str]:
        """Returns cliff style commands with a specific prefix."""
        if not prefix:
            return [n for n, v in self.command_manager]
        return [n for n, v in self.command_manager if n.startswith(prefix)]

    def help_help(self) -> None:
        # Use the command manager to get instructions for "help"
        self.default('help help')

    def do_help(self, arg: str | None) -> None:
        if arg:
            # Check if the arg is a builtin command or something
            # coming from the command manager
            arg_parts = shlex.split(arg)
            method_name = '_'.join(
                itertools.chain(
                    ['do'],
                    itertools.takewhile(
                        lambda x: not x.startswith('-'), arg_parts
                    ),
                )
            )
            # Have the command manager version of the help
            # command produce the help text since cmd and
            # cmd2 do not provide help for "help"
            if hasattr(self, method_name):
                cmd2.Cmd.do_help(self, arg)
                return
            # Dispatch to the underlying help command,
            # which knows how to provide help for extension
            # commands.
            self.default(f'help {arg}')
        else:
            stdout = self.stdout
            try:
                with autopage.argparse.help_pager(stdout) as paged_out:
                    self.stdout = paged_out

                    cmd2.Cmd.do_help(self, arg)  # type: ignore
                    cmd_names = sorted([n for n, v in self.command_manager])
                    self.print_topics(self.app_cmd_header, cmd_names, 15, 80)
            finally:
                self.stdout = stdout
        return

    # Create exit alias to quit the interactive shell.
    do_exit = cmd2.Cmd.do_quit

    def get_names(self) -> list[str]:
        # Override the base class version to filter out
        # things that look like they should be hidden
        # from the user.
        return [
            n for n in cmd2.Cmd.get_names(self) if not n.startswith('do__')
        ]

    def precmd(self, statement: cmd2.Statement | str) -> cmd2.Statement:
        """Hook method executed just before the command is executed by
        :meth:`~cmd2.Cmd.onecmd` and after adding it to history.

        :param statement: subclass of str which also contains the parsed input
        :return: a potentially modified version of the input Statement object
        """
        statement = super().precmd(statement)

        # NOTE(mordred): The above docstring is copied in from cmd2 because
        # current cmd2 has a docstring that sphinx finds if we don't override
        # it, and it breaks sphinx.

        # Pre-process the parsed command in case it looks like one of
        # our subcommands, since cmd2 does not handle multi-part
        # command names by default.
        line_parts = self._split_line(statement)
        try:
            the_cmd = self.command_manager.find_command(line_parts)
            cmd_factory, cmd_name, sub_argv = the_cmd
        except ValueError:
            # Not a plugin command
            pass
        else:
            statement = cmd2.Statement(
                ' '.join(sub_argv),
                raw=statement.raw,
                command=cmd_name,
                arg_list=sub_argv,
                multiline_command=statement.multiline_command,
                terminator=statement.terminator,
                suffix=statement.suffix,
                pipe_to=statement.pipe_to,
                output=statement.output,
                output_to=statement.output_to,
            )
        return statement

    def cmdloop(self, intro: str | None = None) -> None:  # type: ignore[override]
        # We don't want the cmd2 cmdloop() behaviour, just call the old one
        # directly.  In part this is because cmd2.cmdloop() doe not return
        # anything useful and we want to have a useful exit code.
        # FIXME(stephenfin): The above is no longer true: _cmdloop now returns
        # nothing.
        self._cmdloop()



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

"""Output formatters for JSON."""

import argparse
from collections.abc import Iterable, Sequence
import json
from typing import Any, TextIO

from cliff import columns
from cliff.formatters import base


class JSONFormatter(base.ListFormatter, base.SingleFormatter):
    def add_argument_group(self, parser: argparse.ArgumentParser) -> None:
        group = parser.add_argument_group(title='json formatter')
        group.add_argument(
            '--noindent',
            action='store_true',
            dest='noindent',
            help='whether to disable indenting the JSON',
        )

    def emit_list(
        self,
        column_names: Sequence[str],
        data: Iterable[Sequence[Any]],
        stdout: TextIO,
        parsed_args: argparse.Namespace,
    ) -> None:
        items = []
        for item in data:
            items.append(
                {
                    n: (
                        i.machine_readable()
                        if isinstance(i, columns.FormattableColumn)
                        else i
                    )
                    for n, i in zip(column_names, item)
                }
            )
        indent = None if parsed_args.noindent else 2
        json.dump(items, stdout, indent=indent)
        stdout.write('\n')

    def emit_one(
        self,
        column_names: Sequence[str],
        data: Sequence[Any],
        stdout: TextIO,
        parsed_args: argparse.Namespace,
    ) -> None:
        one = {
            n: (
                i.machine_readable()
                if isinstance(i, columns.FormattableColumn)
                else i
            )
            for n, i in zip(column_names, data)
        }
        indent = None if parsed_args.noindent else 2
        json.dump(one, stdout, indent=indent)
        stdout.write('\n')



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

"""Output formatters using csv format."""

import argparse
from collections.abc import Iterable, Sequence
import csv
import os
from typing import Any, Literal, TextIO

from cliff import columns
from cliff.formatters import base


class CSVLister(base.ListFormatter):
    QUOTE_MODES: dict[str, Literal[0, 1, 2, 3]] = {
        'all': csv.QUOTE_ALL,
        'minimal': csv.QUOTE_MINIMAL,
        'nonnumeric': csv.QUOTE_NONNUMERIC,
        'none': csv.QUOTE_NONE,
    }

    def add_argument_group(self, parser: argparse.ArgumentParser) -> None:
        group = parser.add_argument_group('CSV Formatter')
        group.add_argument(
            '--quote',
            choices=sorted(self.QUOTE_MODES.keys()),
            dest='quote_mode',
            default='nonnumeric',
            help='when to include quotes, defaults to nonnumeric',
        )

    def emit_list(
        self,
        column_names: Sequence[str],
        data: Iterable[Sequence[Any]],
        stdout: TextIO,
        parsed_args: argparse.Namespace,
    ) -> None:
        writer = csv.writer(
            stdout,
            quoting=self.QUOTE_MODES[parsed_args.quote_mode],
            lineterminator=os.linesep,
            escapechar='\\',
        )
        writer.writerow(column_names)
        for row in data:
            writer.writerow(
                [
                    (
                        str(c.machine_readable())
                        if isinstance(c, columns.FormattableColumn)
                        else c
                    )
                    for c in row
                ]
            )
        return



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import io
import re

from unittest import mock

from cliff import app as application
from cliff import commandmanager
from cliff import help
from cliff.tests import base
from cliff.tests import utils


class TestHelp(base.TestBase):
    def test_show_help_for_command(self):
        # FIXME(dhellmann): Are commands tied too closely to the app? Or
        # do commands know too much about apps by using them to get to the
        # command manager?
        stdout = io.StringIO()
        app = application.App(
            'testing',
            '1',
            utils.TestCommandManager(utils.TEST_NAMESPACE),
            stdout=stdout,
        )
        app.NAME = 'test'
        help_cmd = help.HelpCommand(app, mock.Mock())
        parser = help_cmd.get_parser('test')
        parsed_args = parser.parse_args(['one'])
        try:
            help_cmd.run(parsed_args)
        except help.HelpExit:
            pass
        self.assertEqual('TestParser', stdout.getvalue())

    def test_list_matching_commands(self):
        # FIXME(dhellmann): Are commands tied too closely to the app? Or
        # do commands know too much about apps by using them to get to the
        # command manager?
        stdout = io.StringIO()
        app = application.App(
            'testing',
            '1',
            utils.TestCommandManager(utils.TEST_NAMESPACE),
            stdout=stdout,
        )
        app.NAME = 'test'
        help_cmd = help.HelpCommand(app, mock.Mock())
        parser = help_cmd.get_parser('test')
        parsed_args = parser.parse_args(['t'])
        try:
            help_cmd.run(parsed_args)
        except help.HelpExit:
            pass
        help_output = stdout.getvalue()
        self.assertIn('Command "t" matches:', help_output)
        self.assertIn('three word command\n  two words\n', help_output)

    def test_list_matching_commands_no_match(self):
        # FIXME(dhellmann): Are commands tied too closely to the app? Or
        # do commands know too much about apps by using them to get to the
        # command manager?
        stdout = io.StringIO()
        app = application.App(
            'testing',
            '1',
            utils.TestCommandManager(utils.TEST_NAMESPACE),
            stdout=stdout,
        )
        app.NAME = 'test'
        help_cmd = help.HelpCommand(app, mock.Mock())
        parser = help_cmd.get_parser('test')
        parsed_args = parser.parse_args(['z'])
        self.assertRaises(
            ValueError,
            help_cmd.run,
            parsed_args,
        )

    def test_show_help_for_help(self):
        # FIXME(dhellmann): Are commands tied too closely to the app? Or
        # do commands know too much about apps by using them to get to the
        # command manager?
        stdout = io.StringIO()
        app = application.App(
            'testing',
            '1',
            utils.TestCommandManager(utils.TEST_NAMESPACE),
            stdout=stdout,
        )
        app.NAME = 'test'
        app.options = mock.Mock()
        help_cmd = help.HelpCommand(app, mock.Mock())
        parser = help_cmd.get_parser('test')
        parsed_args = parser.parse_args([])
        try:
            help_cmd.run(parsed_args)
        except help.HelpExit:
            pass
        help_text = stdout.getvalue()
        usage_line = help_text.splitlines()[0]
        self.assertTrue(usage_line.startswith('usage: '), usage_line)
        self.assertIn('[--version]', usage_line)
        self.assertRegex(help_text, 'option(s|al arguments):\n  --version')
        expected = (
            '  one            Test command\n'
            '  three word command  Test command\n'
        )
        self.assertIn(expected, help_text)

    def test_list_deprecated_commands(self):
        # FIXME(dhellmann): Are commands tied too closely to the app? Or
        # do commands know too much about apps by using them to get to the
        # command manager?
        stdout = io.StringIO()
        app = application.App(
            'testing',
            '1',
            utils.TestCommandManager(utils.TEST_NAMESPACE),
            stdout=stdout,
        )
        app.NAME = 'test'
        try:
            app.run(['--help'])
        except help.HelpExit:
            pass
        help_output = stdout.getvalue()
        self.assertIn('two words', help_output)
        self.assertIn('three word command', help_output)
        self.assertNotIn('old cmd', help_output)

    @mock.patch.object(
        commandmanager.EntryPointWrapper,
        'load',
        side_effect=Exception('Could not load EntryPoint'),
    )
    def test_show_help_with_ep_load_fail(self, mock_load):
        stdout = io.StringIO()
        app = application.App(
            'testing',
            '1',
            utils.TestCommandManager(utils.TEST_NAMESPACE),
            stdout=stdout,
        )
        app.NAME = 'test'
        app.options = mock.Mock()
        app.options.debug = False
        help_cmd = help.HelpCommand(app, mock.Mock())
        parser = help_cmd.get_parser('test')
        parsed_args = parser.parse_args([])
        try:
            help_cmd.run(parsed_args)
        except help.HelpExit:
            pass
        help_output = stdout.getvalue()
        self.assertIn('Commands:', help_output)
        self.assertIn('Could not load', help_output)
        self.assertNotIn('Exception: Could not load EntryPoint', help_output)

    @mock.patch.object(
        commandmanager.EntryPointWrapper,
        'load',
        side_effect=Exception('Could not load EntryPoint'),
    )
    def test_show_help_print_exc_with_ep_load_fail(self, mock_load):
        stdout = io.StringIO()
        app = application.App(
            'testing',
            '1',
            utils.TestCommandManager(utils.TEST_NAMESPACE),
            stdout=stdout,
        )
        app.NAME = 'test'
        app.options = mock.Mock()
        app.options.debug = True
        help_cmd = help.HelpCommand(app, mock.Mock())
        parser = help_cmd.get_parser('test')
        parsed_args = parser.parse_args([])
        try:
            help_cmd.run(parsed_args)
        except help.HelpExit:
            pass
        help_output = stdout.getvalue()
        self.assertIn('Commands:', help_output)
        self.assertIn('Could not load', help_output)
        self.assertIn('Exception: Could not load EntryPoint', help_output)

    def test_show_help_color_output_equivalent_to_plain(self):
        """Colored help should be identical save for ANSI codes."""

        def run_help(color):
            output = io.StringIO()
            pager = mock.MagicMock()
            pager.to_terminal.return_value = color
            pager.__enter__ = mock.Mock(return_value=output)
            pager.__exit__ = mock.Mock(return_value=False)
            app = application.App(
                'testing',
                '1',
                utils.TestCommandManager(utils.TEST_NAMESPACE),
                stdout=io.StringIO(),
            )
            app.NAME = 'test'
            with mock.patch(
                'cliff.help.autopage.argparse.help_pager', return_value=pager
            ):
                with mock.patch(
                    'cliff.help.autopage.argparse.use_color_for_parser'
                ):
                    try:
                        app.run(['--help'])
                    except help.HelpExit:
                        pass
            return output.getvalue()

        plain = run_help(color=False)
        colored = run_help(color=True)
        self.assertNotEqual(plain, colored)
        self.assertEqual(plain, re.sub(r'\x1b\[[0-9;]*m', '', colored))



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

"""Bash completion tests"""

import io
from unittest import mock

from cliff import app as application
from cliff import commandmanager
from cliff import complete
from cliff.tests import base


class TestCompletion(base.TestBase):
    def test_dictionary(self):
        sot = complete.CompleteDictionary()
        sot.add_command(
            "image delete".split(), [mock.Mock(option_strings=["1"])]
        )
        sot.add_command(
            "image list".split(), [mock.Mock(option_strings=["2"])]
        )
        sot.add_command(
            "image create".split(), [mock.Mock(option_strings=["3"])]
        )
        sot.add_command(
            "volume type create".split(), [mock.Mock(option_strings=["4"])]
        )
        sot.add_command(
            "volume type delete".split(), [mock.Mock(option_strings=["5"])]
        )
        self.assertEqual("image volume", sot.get_commands())
        result = sot.get_data()
        self.assertEqual("image", result[0][0])
        self.assertEqual("create delete list", result[0][1])
        self.assertEqual("image_create", result[1][0])
        self.assertEqual("3", result[1][1])
        self.assertEqual("image_delete", result[2][0])
        self.assertEqual("1", result[2][1])
        self.assertEqual("image_list", result[3][0])
        self.assertEqual("2", result[3][1])

    def test_complete_dictionary_subcmd(self):
        sot = complete.CompleteDictionary()
        sot.add_command(
            "image delete".split(), [mock.Mock(option_strings=["1"])]
        )
        sot.add_command(
            "image list".split(), [mock.Mock(option_strings=["2"])]
        )
        sot.add_command(
            "image list better".split(), [mock.Mock(option_strings=["3"])]
        )
        self.assertEqual("image", sot.get_commands())
        result = sot.get_data()
        self.assertEqual("image", result[0][0])
        self.assertEqual("delete list list_better", result[0][1])
        self.assertEqual("image_delete", result[1][0])
        self.assertEqual("1", result[1][1])
        self.assertEqual("image_list", result[2][0])
        self.assertEqual("2 better", result[2][1])
        self.assertEqual("image_list_better", result[3][0])
        self.assertEqual("3", result[3][1])


class TestCompletionAlternatives(base.TestBase):
    def given_cmdo_data(self):
        cmdo = "image server"
        data = [
            ("image", "create"),
            ("image_create", "--eolus"),
            ("server", "meta ssh"),
            ("server_meta_delete", "--wilson"),
            ("server_ssh", "--sunlight"),
        ]
        return cmdo, data

    def then_data(self, content):
        self.assertIn("  cmds='image server'", content)
        self.assertIn("  cmds_image='create'", content)
        self.assertIn("  cmds_image_create='--eolus'", content)
        self.assertIn("  cmds_server='meta ssh'", content)
        self.assertIn("  cmds_server_meta_delete='--wilson'", content)
        self.assertIn("  cmds_server_ssh='--sunlight'", content)

    def test_complete_no_code(self):
        output = io.StringIO()
        sot = complete.CompleteNoCode("doesNotMatter", output)
        sot.write(*self.given_cmdo_data())
        self.then_data(output.getvalue().splitlines())

    def test_complete_bash(self):
        output = io.StringIO()
        sot = complete.CompleteBash("openstack", output)
        sot.write(*self.given_cmdo_data())
        self.then_data(output.getvalue().splitlines())
        self.assertIn("_openstack()", output.getvalue().splitlines()[0])
        self.assertIn(
            "complete -F _openstack openstack",
            output.getvalue().splitlines()[-1],
        )

    def test_complete_command_parser(self):
        sot = complete.CompleteCommand(mock.Mock(), mock.Mock())
        parser = sot.get_parser('nothing')
        self.assertEqual("nothing", parser.prog)
        self.assertEqual("print bash completion command", parser.description)


class TestCompletionAction(base.TestBase):
    def given_complete_command(self):
        cmd_mgr = commandmanager.CommandManager('cliff.tests')
        app = application.App('testing', '1', cmd_mgr, stdout=io.StringIO())
        sot = complete.CompleteCommand(app, mock.Mock())
        cmd_mgr.add_command('complete', complete.CompleteCommand)
        return sot, app, cmd_mgr

    def then_actions_equal(self, actions):
        optstr = ' '.join(
            opt for action in actions for opt in action.option_strings
        )
        self.assertEqual('-h --help --name --shell', optstr)

    def test_complete_command_get_actions(self):
        sot, app, cmd_mgr = self.given_complete_command()
        app.interactive_mode = False
        actions = sot.get_actions(["complete"])
        self.then_actions_equal(actions)

    def test_complete_command_get_actions_interactive(self):
        sot, app, cmd_mgr = self.given_complete_command()
        app.interactive_mode = True
        actions = sot.get_actions(["complete"])
        self.then_actions_equal(actions)

    def test_complete_command_take_action(self):
        sot, app, cmd_mgr = self.given_complete_command()
        parsed_args = mock.Mock()
        parsed_args.name = "test_take"
        parsed_args.shell = "bash"
        self.assertEqual(0, sot.take_action(parsed_args))
        content = app.stdout.getvalue().splitlines()
        self.assertIn("_test_take()", content[0])
        self.assertIn("complete -F _test_take test_take", content[-1])
        self.assertIn("  cmds='complete help'", content)
        self.assertIn("  cmds_complete='-h --help --name --shell'", content)
        self.assertIn("  cmds_help='-h --help'", content)

    def test_complete_command_remove_dashes(self):
        sot, app, cmd_mgr = self.given_complete_command()
        parsed_args = mock.Mock()
        parsed_args.name = "test-take"
        parsed_args.shell = "bash"
        self.assertEqual(0, sot.take_action(parsed_args))
        content = app.stdout.getvalue().splitlines()
        self.assertIn("_test_take()", content[0])
        self.assertIn("complete -F _test_take test-take", content[-1])



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import argparse
import io
import unittest
from unittest import mock

from cliff.formatters import commaseparated
from cliff.tests import test_columns


class TestCSVFormatter(unittest.TestCase):
    def test_commaseparated_list_formatter(self):
        sf = commaseparated.CSVLister()
        c = ('a', 'b', 'c')
        d1 = ('A', 'B', 'C')
        d2 = ('D', 'E', 'F')
        data = [d1, d2]
        expected = 'a,b,c\nA,B,C\nD,E,F\n'
        output = io.StringIO()
        parsed_args = mock.Mock()
        parsed_args.quote_mode = 'none'
        sf.emit_list(c, data, output, parsed_args)
        actual = output.getvalue()
        self.assertEqual(expected, actual)

    def test_commaseparated_list_formatter_quoted(self):
        sf = commaseparated.CSVLister()
        c = ('a', 'b', 'c')
        d1 = ('A', 'B', 'C')
        d2 = ('D', 'E', 'F')
        data = [d1, d2]
        expected = '"a","b","c"\n"A","B","C"\n"D","E","F"\n'
        output = io.StringIO()
        # Parse arguments as if passed on the command-line
        parser = argparse.ArgumentParser(description='Testing...')
        sf.add_argument_group(parser)
        parsed_args = parser.parse_args(['--quote', 'all'])
        sf.emit_list(c, data, output, parsed_args)
        actual = output.getvalue()
        self.assertEqual(expected, actual)

    def test_commaseparated_list_formatter_formattable_column(self):
        sf = commaseparated.CSVLister()
        c = ('a', 'b', 'c')
        d1 = ('A', 'B', test_columns.FauxColumn(['the', 'value']))
        data = [d1]
        expected = 'a,b,c\nA,B,[\'the\'\\, \'value\']\n'
        output = io.StringIO()
        parsed_args = mock.Mock()
        parsed_args.quote_mode = 'none'
        sf.emit_list(c, data, output, parsed_args)
        actual = output.getvalue()
        self.assertEqual(expected, actual)

    def test_commaseparated_list_formatter_unicode(self):
        sf = commaseparated.CSVLister()
        c = ('a', 'b', 'c')
        d1 = ('A', 'B', 'C')
        happy = '高兴'
        d2 = ('D', 'E', happy)
        data = [d1, d2]
        expected = f'a,b,c\nA,B,C\nD,E,{happy}\n'
        output = io.StringIO()
        parsed_args = mock.Mock()
        parsed_args.quote_mode = 'none'
        sf.emit_list(c, data, output, parsed_args)
        actual = output.getvalue()
        self.assertEqual(expected, actual)



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

"""Base classes for formatters."""

import abc
import argparse
from collections.abc import Iterable, Sequence
from typing import Any, TextIO, TypeVar


FormatterT = TypeVar('FormatterT', bound='Formatter')


class Formatter(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def add_argument_group(self, parser: argparse.ArgumentParser) -> None:
        """Add any options to the argument parser.

        Should use our own argument group.
        """


class ListFormatter(Formatter, metaclass=abc.ABCMeta):
    """Base class for formatters that know how to deal with many objects."""

    @abc.abstractmethod
    def emit_list(
        self,
        column_names: Sequence[str],
        data: Iterable[Sequence[Any]],
        stdout: TextIO,
        parsed_args: argparse.Namespace,
    ) -> None:
        """Format and print the list from the iterable data source.

        Data values can be primitive types like ints and strings, or
        can be an instance of a :class:`FormattableColumn` for
        situations where the value is complex, and may need to be
        handled differently for human readable output vs. machine
        readable output.

        :param column_names: names of the columns
        :param data: iterable data source, one tuple per object
                     with values in order of column names
        :param stdout: output stream where data should be written
        :param parsed_args: argparse namespace from our local options
        """


class SingleFormatter(Formatter, metaclass=abc.ABCMeta):
    """Base class for formatters that work with single objects."""

    @abc.abstractmethod
    def emit_one(
        self,
        column_names: Sequence[str],
        data: Sequence[Any],
        stdout: TextIO,
        parsed_args: argparse.Namespace,
    ) -> None:
        """Format and print the values associated with the single object.

        Data values can be primitive types like ints and strings, or
        can be an instance of a :class:`FormattableColumn` for
        situations where the value is complex, and may need to be
        handled differently for human readable output vs. machine
        readable output.

        :param column_names: names of the columns
        :param data: iterable data source with values in order of column names
        :param stdout: output stream where data should be written
        :param parsed_args: argparse namespace from our local options
        """



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

"""Output formatters values only"""

import argparse
from collections.abc import Iterable, Sequence
from typing import Any, TextIO

from cliff import columns
from cliff.formatters import base


class ValueFormatter(base.ListFormatter, base.SingleFormatter):
    def add_argument_group(self, parser: argparse.ArgumentParser) -> None:
        pass

    def emit_list(
        self,
        column_names: Sequence[str],
        data: Iterable[Sequence[Any]],
        stdout: TextIO,
        parsed_args: argparse.Namespace,
    ) -> None:
        for row in data:
            stdout.write(
                ' '.join(
                    str(
                        c.machine_readable()
                        if isinstance(c, columns.FormattableColumn)
                        else c
                    )
                    for c in row
                )
                + '\n'
            )
        return

    def emit_one(
        self,
        column_names: Sequence[str],
        data: Sequence[Any],
        stdout: TextIO,
        parsed_args: argparse.Namespace,
    ) -> None:
        for value in data:
            stdout.write(
                '{}\n'.format(
                    str(
                        value.machine_readable()
                        if isinstance(value, columns.FormattableColumn)
                        else value
                    )
                )
            )
        return



#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.


from cliff import app
from cliff.interactive import InteractiveApp
from cliff.tests import base
from cliff.tests import utils


class TestInteractive(base.TestBase):
    def make_interactive_app(self, errexit, *command_names):
        fake_command_manager = [(x, None) for x in command_names]
        return InteractiveApp(
            app.App(
                'foo', '1.0', utils.TestCommandManager(utils.TEST_NAMESPACE)
            ),
            fake_command_manager,  # type: ignore
            stdin=None,
            stdout=None,
            errexit=errexit,
        )

    def _test_completedefault(self, expecteds, line, begidx):
        command_names = set(
            ['show file', 'show folder', 'show  long', 'list all']
        )
        app = self.make_interactive_app(False, *command_names)
        observeds = app.completedefault(None, line, begidx, None)
        self.assertEqual(set(expecteds), set(observeds))
        self.assertTrue(
            set([line[:begidx] + x for x in observeds]) <= command_names
        )

    def test_empty_text_completedefault(self):
        # line = 'show ' + begidx = 5 implies text = ''
        self._test_completedefault(['file', 'folder', ' long'], 'show ', 5)

    def test_nonempty_text_completedefault2(self):
        # line = 'show f' + begidx = 6 implies text = 'f'
        self._test_completedefault(['file', 'folder'], 'show f', 5)

    def test_long_completedefault(self):
        self._test_completedefault(['long'], 'show  ', 6)

    def test_no_completedefault(self):
        self._test_completedefault([], 'taz ', 4)

    def test_no_errexit(self):
        command_names = set(['show file', 'show folder', 'list all'])
        app = self.make_interactive_app(False, *command_names)
        self.assertFalse(app.errexit)

    def test_errexit(self):
        command_names = set(['show file', 'show folder', 'list all'])
        app = self.make_interactive_app(True, *command_names)
        self.assertTrue(app.errexit)



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import argparse
import importlib.metadata

from cliff import app as application
from cliff import command
from cliff import commandmanager
from cliff import hooks
from cliff import lister
from cliff import show
from cliff.tests import base

from stevedore import extension
from unittest import mock


def make_extension(name, plugin, obj):
    ep = importlib.metadata.EntryPoint(
        name=name,
        value=f'my_module:{name.replace("-", "_")}',
        group='my_app.plugins',
    )
    return extension.Extension(name, ep, plugin, obj)


def make_app(**kwargs):
    cmd_mgr = commandmanager.CommandManager('cliff.tests')

    # Register a command that succeeds
    cmd = mock.MagicMock(spec=command.Command)
    command_inst = mock.MagicMock(spec=command.Command)
    command_inst.run.return_value = 0
    cmd.return_value = command_inst
    cmd_mgr.add_command('mock', cmd)

    # Register a command that fails
    err_command = mock.Mock(name='err_command', spec=command.Command)
    err_command_inst = mock.Mock(spec=command.Command)
    err_command_inst.run = mock.Mock(
        side_effect=RuntimeError('test exception')
    )
    err_command.return_value = err_command_inst
    cmd_mgr.add_command('error', err_command)

    app = application.App(
        'testing command hooks',
        '1',
        cmd_mgr,
        stderr=mock.Mock(),  # suppress warning messages
        **kwargs,
    )
    return app


class TestCommand(command.Command):
    """Description of command."""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        return parser

    def take_action(self, parsed_args):
        return 42


class TestShowCommand(show.ShowOne):
    """Description of command."""

    def take_action(self, parsed_args):
        return (('Name',), ('value',))


class TestListerCommand(lister.Lister):
    """Description of command."""

    def take_action(self, parsed_args):
        return (('Name',), [('value',)])


class TestHook(hooks.CommandHook):
    """A hook that returns an epilog and a new argument."""

    _before_called = False
    _after_called = False

    def get_parser(self, parser):
        parser.add_argument('--added-by-hook')
        return parser

    def get_epilog(self):
        return 'hook epilog'

    def before(self, parsed_args):
        self._before_called = True

    def after(self, parsed_args, return_code):
        self._after_called = True


class TestNullHook(hooks.CommandHook):
    """A hook that does not return an epilog nor any new arguments."""

    _before_called = False
    _after_called = False

    def get_parser(self, parser):
        return None

    def get_epilog(self):
        return None

    def before(self, parsed_args):
        self._before_called = True

    def after(self, parsed_args, return_code):
        self._after_called = True


class TestChangeHook(hooks.CommandHook):
    """A hook that modifies the parsed arguments and the return code.

    This would be used with subclasses of the Command base command class.
    """

    _before_called = False
    _after_called = False

    def get_parser(self, parser):
        parser.add_argument('--added-by-hook')
        return parser

    def get_epilog(self):
        return 'hook epilog'

    def before(self, parsed_args):
        self._before_called = True
        parsed_args.added_by_hook = 'othervalue'
        parsed_args.added_by_before = True
        return parsed_args

    def after(self, parsed_args, return_code):
        self._after_called = True
        return 24


class TestDisplayChangeHook(hooks.CommandHook):
    """A hook that modifies the parsed arguments and the output.

    This would be used with subclasses of :class:`cliff.show.ShowOne`.
    """

    _before_called = False
    _after_called = False

    def get_parser(self, parser):
        parser.add_argument('--added-by-hook')
        return parser

    def get_epilog(self):
        return 'hook epilog'

    def before(self, parsed_args):
        self._before_called = True
        parsed_args.added_by_hook = 'othervalue'
        parsed_args.added_by_before = True
        return parsed_args

    def after(self, parsed_args, return_code):
        self._after_called = True
        return (('Name',), ('othervalue',))


class TestListerChangeHook(hooks.CommandHook):
    """A hook that modifies the parsed arguments and the output.

    This would be used with subclasses of :class:`cliff.show.Lister`.
    """

    _before_called = False
    _after_called = False

    def get_parser(self, parser):
        parser.add_argument('--added-by-hook')
        return parser

    def get_epilog(self):
        return 'hook epilog'

    def before(self, parsed_args):
        self._before_called = True
        parsed_args.added_by_hook = 'othervalue'
        parsed_args.added_by_before = True
        return parsed_args

    def after(self, parsed_args, return_code):
        self._after_called = True
        return (('Name',), [('othervalue',)])


class TestCommandLoadHooks(base.TestBase):
    def test_no_name(self):
        app = make_app()
        cmd = TestCommand(app, None)
        self.assertEqual([], cmd._hooks)

    @mock.patch('stevedore.extension.ExtensionManager')
    def test_app_and_name(self, em):
        app = make_app()
        TestCommand(app, None, cmd_name='test')
        print(em.mock_calls[0])
        name, args, kwargs = em.mock_calls[0]
        print(kwargs)
        self.assertEqual('cliff.tests.test', kwargs['namespace'])


class TestHooks(base.TestBase):
    def setUp(self):
        super().setUp()
        self.app = make_app()
        self.cmd = TestCommand(self.app, None, cmd_name='test')
        self.hook_a = TestHook(self.cmd)
        self.hook_b = TestNullHook(self.cmd)
        self.mgr: extension.ExtensionManager[hooks.CommandHook]
        self.mgr = extension.ExtensionManager.make_test_instance(
            [
                make_extension('parser-hook-a', TestHook, self.hook_a),
                make_extension('parser-hook-b', TestHook, self.hook_b),
            ],
        )
        # Replace the auto-loaded hooks with our explicitly created
        # manager.
        self.cmd._hooks = self.mgr

    def test_get_parser(self):
        parser = self.cmd.get_parser('test')
        results = parser.parse_args(['--added-by-hook', 'value'])
        self.assertEqual(results.added_by_hook, 'value')

    def test_get_epilog(self):
        results = self.cmd.get_epilog()
        self.assertIn('hook epilog', results)

    def test_before(self):
        self.assertFalse(self.hook_a._before_called)
        self.assertFalse(self.hook_b._before_called)
        self.cmd.run(argparse.Namespace())
        self.assertTrue(self.hook_a._before_called)
        self.assertTrue(self.hook_b._before_called)

    def test_after(self):
        self.assertFalse(self.hook_a._after_called)
        self.assertFalse(self.hook_b._after_called)
        result = self.cmd.run(argparse.Namespace())
        self.assertTrue(self.hook_a._after_called)
        self.assertTrue(self.hook_b._after_called)
        self.assertEqual(result, 42)


class TestChangeHooks(base.TestBase):
    def setUp(self):
        super().setUp()
        self.app = make_app()
        self.cmd = TestCommand(self.app, None, cmd_name='test')
        self.hook = TestChangeHook(self.cmd)
        self.mgr: extension.ExtensionManager[hooks.CommandHook]
        self.mgr = extension.ExtensionManager.make_test_instance(
            [make_extension('parser-hook', TestChangeHook, self.hook)]
        )
        # Replace the auto-loaded hooks with our explicitly created
        # manager.
        self.cmd._hooks = self.mgr

    def test_get_parser(self):
        parser = self.cmd.get_parser('test')
        results = parser.parse_args(['--added-by-hook', 'value'])
        self.assertEqual(results.added_by_hook, 'value')

    def test_get_epilog(self):
        results = self.cmd.get_epilog()
        self.assertIn('hook epilog', results)

    def test_before(self):
        self.assertFalse(self.hook._before_called)
        parser = self.cmd.get_parser('test')
        results = parser.parse_args(['--added-by-hook', 'value'])
        self.cmd.run(results)
        self.assertTrue(self.hook._before_called)
        self.assertEqual(results.added_by_hook, 'othervalue')
        self.assertTrue(results.added_by_before)

    def test_after(self):
        self.assertFalse(self.hook._after_called)
        parser = self.cmd.get_parser('test')
        results = parser.parse_args(['--added-by-hook', 'value'])
        result = self.cmd.run(results)
        self.assertTrue(self.hook._after_called)
        self.assertEqual(result, 24)


class TestShowOneHooks(base.TestBase):
    def setUp(self):
        super().setUp()
        self.app = make_app()
        self.cmd = TestShowCommand(self.app, None, cmd_name='test')
        self.hook = TestHook(self.cmd)
        self.mgr: extension.ExtensionManager[hooks.CommandHook]
        self.mgr = extension.ExtensionManager.make_test_instance(
            [make_extension('parser-hook', TestHook, self.hook)]
        )
        # Replace the auto-loaded hooks with our explicitly created
        # manager.
        self.cmd._hooks = self.mgr

    def test_get_parser(self):
        parser = self.cmd.get_parser('test')
        results = parser.parse_args(['--added-by-hook', 'value'])
        self.assertEqual(results.added_by_hook, 'value')

    def test_get_epilog(self):
        results = self.cmd.get_epilog()
        self.assertIn('hook epilog', results)

    def test_before(self):
        self.assertFalse(self.hook._before_called)
        parser = self.cmd.get_parser('test')
        results = parser.parse_args(['--added-by-hook', 'value'])
        self.cmd.run(results)
        self.assertTrue(self.hook._before_called)

    def test_after(self):
        self.assertFalse(self.hook._after_called)
        parser = self.cmd.get_parser('test')
        results = parser.parse_args(['--added-by-hook', 'value'])
        self.cmd.run(results)
        self.assertTrue(self.hook._after_called)


class TestShowOneChangeHooks(base.TestBase):
    def setUp(self):
        super().setUp()
        self.app = make_app()
        self.cmd = TestShowCommand(self.app, None, cmd_name='test')
        self.hook = TestDisplayChangeHook(self.cmd)
        self.mgr: extension.ExtensionManager[hooks.CommandHook]
        self.mgr = extension.ExtensionManager.make_test_instance(
            [make_extension('parser-hook', TestDisplayChangeHook, self.hook)]
        )
        # Replace the auto-loaded hooks with our explicitly created
        # manager.
        self.cmd._hooks = self.mgr

    def test_get_parser(self):
        parser = self.cmd.get_parser('test')
        results = parser.parse_args(['--added-by-hook', 'value'])
        self.assertEqual(results.added_by_hook, 'value')

    def test_get_epilog(self):
        results = self.cmd.get_epilog()
        self.assertIn('hook epilog', results)

    def test_before(self):
        self.assertFalse(self.hook._before_called)
        parser = self.cmd.get_parser('test')
        results = parser.parse_args(['--added-by-hook', 'value'])
        self.cmd.run(results)
        self.assertTrue(self.hook._before_called)
        self.assertEqual(results.added_by_hook, 'othervalue')
        self.assertTrue(results.added_by_before)

    def test_after(self):
        self.assertFalse(self.hook._after_called)
        parser = self.cmd.get_parser('test')
        results = parser.parse_args(['--added-by-hook', 'value'])
        result = self.cmd.run(results)
        self.assertTrue(self.hook._after_called)
        self.assertEqual(result, 0)


class TestListerHooks(base.TestBase):
    def setUp(self):
        super().setUp()
        self.app = make_app()
        self.cmd = TestListerCommand(self.app, None, cmd_name='test')
        self.hook = TestHook(self.cmd)
        self.mgr: extension.ExtensionManager[hooks.CommandHook]
        self.mgr = extension.ExtensionManager.make_test_instance(
            [make_extension('parser-hook', TestHook, self.hook)]
        )
        # Replace the auto-loaded hooks with our explicitly created
        # manager.
        self.cmd._hooks = self.mgr

    def test_get_parser(self):
        parser = self.cmd.get_parser('test')
        results = parser.parse_args(['--added-by-hook', 'value'])
        self.assertEqual(results.added_by_hook, 'value')

    def test_get_epilog(self):
        results = self.cmd.get_epilog()
        self.assertIn('hook epilog', results)

    def test_before(self):
        self.assertFalse(self.hook._before_called)
        parser = self.cmd.get_parser('test')
        results = parser.parse_args(['--added-by-hook', 'value'])
        self.cmd.run(results)
        self.assertTrue(self.hook._before_called)

    def test_after(self):
        self.assertFalse(self.hook._after_called)
        parser = self.cmd.get_parser('test')
        results = parser.parse_args(['--added-by-hook', 'value'])
        self.cmd.run(results)
        self.assertTrue(self.hook._after_called)


class TestListerChangeHooks(base.TestBase):
    def setUp(self):
        super().setUp()
        self.app = make_app()
        self.cmd = TestListerCommand(self.app, None, cmd_name='test')
        self.hook = TestListerChangeHook(self.cmd)
        self.mgr: extension.ExtensionManager[hooks.CommandHook]
        self.mgr = extension.ExtensionManager.make_test_instance(
            [make_extension('parser-hook', TestListerChangeHook, self.hook)]
        )
        # Replace the auto-loaded hooks with our explicitly created
        # manager.
        self.cmd._hooks = self.mgr

    def test_get_parser(self):
        parser = self.cmd.get_parser('test')
        results = parser.parse_args(['--added-by-hook', 'value'])
        self.assertEqual(results.added_by_hook, 'value')

    def test_get_epilog(self):
        results = self.cmd.get_epilog()
        self.assertIn('hook epilog', results)

    def test_before(self):
        self.assertFalse(self.hook._before_called)
        parser = self.cmd.get_parser('test')
        results = parser.parse_args(['--added-by-hook', 'value'])
        self.cmd.run(results)
        self.assertTrue(self.hook._before_called)
        self.assertEqual(results.added_by_hook, 'othervalue')
        self.assertTrue(results.added_by_before)

    def test_after(self):
        self.assertFalse(self.hook._after_called)
        parser = self.cmd.get_parser('test')
        results = parser.parse_args(['--added-by-hook', 'value'])
        result = self.cmd.run(results)
        self.assertTrue(self.hook._after_called)
        self.assertEqual(result, 0)



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from cliff.command import Command
from cliff.commandmanager import CommandManager

TEST_NAMESPACE = 'cliff.test'


class TestParser:
    def print_help(self, stdout):
        stdout.write('TestParser')


class TestCommand(Command):
    "Test command."

    def get_parser(self, ignore):
        # Make it look like this class is the parser
        # so parse_args() is called.
        return TestParser()

    def take_action(self, args):
        return


class TestDeprecatedCommand(TestCommand):
    deprecated = True


class TestCommandManager(CommandManager):
    def load_commands(self, namespace):
        if namespace == TEST_NAMESPACE:
            for key in ('one', 'two words', 'three word command'):
                self.add_command(key, TestCommand)
            self.add_command('old cmd', TestDeprecatedCommand)



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

"""Application base class."""

import argparse
import inspect
import locale
import logging
import logging.handlers
import os
import sys
from typing import TYPE_CHECKING, Any, TextIO

from cliff import _argparse
from . import complete
from . import help
from . import utils

if TYPE_CHECKING:
    from . import command as _command
    from . import commandmanager as _commandmanager
    from . import interactive as _interactive

logging.getLogger('cliff').addHandler(logging.NullHandler())


# Exit code for exiting due to a signal is 128 + the signal number
_SIGINT_EXIT = 130
_SIGPIPE_EXIT = 141


class App:
    """Application base class.

    :param description: one-liner explaining the program purpose
    :paramtype description: str
    :param version: application version number
    :paramtype version: str
    :param command_manager: plugin loader
    :paramtype command_manager: cliff.commandmanager.CommandManager
    :param stdin: Standard input stream
    :paramtype stdin: readable I/O stream
    :param stdout: Standard output stream
    :paramtype stdout: writable I/O stream
    :param stderr: Standard error output stream
    :paramtype stderr: writable I/O stream
    :param interactive_app_factory: callable to create an
                                    interactive application
    :paramtype interactive_app_factory: cliff.interactive.InteractiveApp
    :param deferred_help: True - Allow subcommands to accept --help with
                          allowing to defer help print after initialize_app
    :paramtype deferred_help: bool
    """

    NAME = os.path.splitext(os.path.basename(sys.argv[0]))[0]
    if NAME == '__main__':
        NAME = os.path.split(os.path.dirname(sys.argv[0]))[-1]
    LOG = logging.getLogger(NAME)

    CONSOLE_MESSAGE_FORMAT = '%(message)s'
    LOG_FILE_MESSAGE_FORMAT = (
        '[%(asctime)s] %(levelname)-8s %(name)s %(message)s'
    )
    DEFAULT_VERBOSE_LEVEL = 1
    DEFAULT_OUTPUT_ENCODING = 'utf-8'

    def __init__(
        self,
        description: str | None,
        version: str | None,
        command_manager: '_commandmanager.CommandManager',
        stdin: TextIO | None = None,
        stdout: TextIO | None = None,
        stderr: TextIO | None = None,
        interactive_app_factory: type['_interactive.InteractiveApp']
        | None = None,
        deferred_help: bool = False,
    ) -> None:
        """Initialize the application."""
        self.command_manager = command_manager
        self.command_manager.add_command('help', help.HelpCommand)
        self.command_manager.add_command('complete', complete.CompleteCommand)
        self._set_streams(stdin, stdout, stderr)
        self.interactive_app_factory = interactive_app_factory
        self.deferred_help = deferred_help
        self.parser = self.build_option_parser(description, version)
        self.interactive_mode = False
        self.interpreter: _interactive.InteractiveApp | None = None

    def _set_streams(
        self,
        stdin: TextIO | None,
        stdout: TextIO | None,
        stderr: TextIO | None,
    ) -> None:
        try:
            locale.setlocale(locale.LC_ALL, '')
        except locale.Error:
            pass

        # Unicode must be encoded/decoded for text I/O streams, the
        # correct encoding for the stream must be selected and it must
        # be capable of handling the set of characters in the stream
        # or Python will raise a codec error. The correct codec is
        # selected based on the locale. Python2 uses the locales
        # encoding but only when the I/O stream is attached to a
        # terminal (TTY) otherwise it uses the default ASCII
        # encoding. The effect is internationalized text written to
        # the terminal works as expected but if command line output is
        # redirected (file or pipe) the ASCII codec is used and the
        # program aborts with a codec error.
        #
        # The default I/O streams stdin, stdout and stderr can be
        # wrapped in a codec based on the locale thus assuring the
        # users desired encoding is always used no matter the I/O
        # destination. Python3 does this by default.
        #
        # If the caller supplies an I/O stream we use it unmodified on
        # the assumption the caller has taken all responsibility for
        # the stream.  But with Python2 if the caller allows us to
        # default the I/O streams to sys.stdin, sys.stdout and
        # sys.stderr we apply the locales encoding just as Python3
        # would do. We also check to make sure the main Python program
        # has not already already wrapped sys.stdin, sys.stdout and
        # sys.stderr as this is a common recommendation.

        self.stdin = stdin or sys.stdin
        self.stdout = stdout or sys.stdout
        self.stderr = stderr or sys.stderr

    def build_option_parser(
        self,
        description: str | None,
        version: str | None,
        argparse_kwargs: dict[str, Any] | None = None,
    ) -> argparse.ArgumentParser:
        """Return an argparse option parser for this application.

        Subclasses may override this method to extend
        the parser with more global options.

        :param description: full description of the application
        :paramtype description: str
        :param version: version number for the application
        :paramtype version: str
        :param argparse_kwargs: extra keyword argument passed to the
                                ArgumentParser constructor
        :paramtype extra_kwargs: dict
        """
        argparse_kwargs = argparse_kwargs or {}
        parser = _argparse.ArgumentParser(
            description=description, add_help=False, **argparse_kwargs
        )
        parser.add_argument(
            '--version',
            action='version',
            version=f'{App.NAME} {version}',
        )
        verbose_group = parser.add_mutually_exclusive_group()
        verbose_group.add_argument(
            '-v',
            '--verbose',
            action='count',
            dest='verbose_level',
            default=self.DEFAULT_VERBOSE_LEVEL,
            help='Increase verbosity of output. Can be repeated.',
        )
        verbose_group.add_argument(
            '-q',
            '--quiet',
            action='store_const',
            dest='verbose_level',
            const=0,
            help='Suppress output except warnings and errors.',
        )
        parser.add_argument(
            '--log-file',
            action='store',
            default=None,
            help='Specify a file to log output. Disabled by default.',
        )
        if self.deferred_help:
            parser.add_argument(
                '-h',
                '--help',
                dest='deferred_help',
                action='store_true',
                help="Show help message and exit.",
            )
        else:
            parser.add_argument(
                '-h',
                '--help',
                action=help.HelpAction,
                nargs=0,
                default=self,  # tricky
                help="Show help message and exit.",
            )
        parser.add_argument(
            '--debug',
            default=False,
            action='store_true',
            help='Show tracebacks on errors.',
        )
        return parser

    def configure_logging(self) -> None:
        """Create logging handlers for any log output."""
        root_logger = logging.getLogger('')
        root_logger.setLevel(logging.DEBUG)

        # Set up logging to a file
        if self.options.log_file:
            file_handler = logging.FileHandler(
                filename=self.options.log_file,
            )
            formatter = logging.Formatter(self.LOG_FILE_MESSAGE_FORMAT)
            file_handler.setFormatter(formatter)
            root_logger.addHandler(file_handler)

        # Always send higher-level messages to the console via stderr
        console = logging.StreamHandler(self.stderr)
        console_level = {
            0: logging.WARNING,
            1: logging.INFO,
            2: logging.DEBUG,
        }.get(self.options.verbose_level, logging.DEBUG)
        console.setLevel(console_level)
        formatter = logging.Formatter(self.CONSOLE_MESSAGE_FORMAT)
        console.setFormatter(formatter)
        root_logger.addHandler(console)
        return

    def print_help_if_requested(self) -> None:
        """Print help and exits if deferred help is enabled and requested.

        '--help' shows the help message and exits:
         * without calling initialize_app if not self.deferred_help (default),
         * after initialize_app call if self.deferred_help,
         * during initialize_app call if self.deferred_help and subclass calls
           explicitly this method in initialize_app.
        """
        if self.deferred_help and self.options.deferred_help:
            action = help.HelpAction([], argparse.SUPPRESS, default=self)
            action(self.parser, self.options, None, None)

    def run(self, argv: list[str]) -> int:
        """Equivalent to the main program for the application.

        :param argv: input arguments and options
        :paramtype argv: list of str
        """

        try:
            self.options, remainder = self.parser.parse_known_args(argv)
            self.configure_logging()
            self.interactive_mode = not remainder
            if self.deferred_help and self.options.deferred_help and remainder:
                # When help is requested and `remainder` has any values disable
                # `deferred_help` and instead allow the help subcommand to
                # handle the request during run_subcommand(). This turns
                # "app foo bar --help" into "app help foo bar". However, when
                # `remainder` is empty use print_help_if_requested() to allow
                # for an early exit.
                # Disabling `deferred_help` here also ensures that
                # print_help_if_requested will not fire if called by a subclass
                # during its initialize_app().
                self.options.deferred_help = False
                remainder.insert(0, "help")
            self.initialize_app(remainder)
            self.print_help_if_requested()
        except BrokenPipeError:
            return _SIGPIPE_EXIT
        except Exception as err:
            if hasattr(self, 'options'):
                debug = self.options.debug
            else:
                debug = True
            if debug:
                self.LOG.exception(err)
                raise
            else:
                self.LOG.error(err)
            return 1
        except KeyboardInterrupt:
            return _SIGINT_EXIT
        result = 1
        if self.interactive_mode:
            self.interact()
        else:
            try:
                result = self.run_subcommand(remainder)
            except BrokenPipeError:
                return _SIGPIPE_EXIT
            except KeyboardInterrupt:
                return _SIGINT_EXIT
        return result

    # FIXME(dhellmann): Consider moving these command handling methods
    # to a separate class.
    def initialize_app(self, argv: list[str]) -> None:
        """Hook for subclasses to take global initialization action
        after the arguments are parsed but before a command is run.
        Invoked only once, even in interactive mode.

        :param argv: List of arguments, including the subcommand to run.
                     Empty for interactive mode.
        """
        return

    def prepare_to_run_command(self, cmd: '_command.Command') -> None:
        """Perform any preliminary work needed to run a command.

        :param cmd: command processor being invoked
        :paramtype cmd: cliff.command.Command
        """
        return

    def clean_up(
        self,
        cmd: '_command.Command',
        result: int,
        err: BaseException | None,
    ) -> None:
        """Hook run after a command is done to shutdown the app.

        :param cmd: command processor being invoked
        :paramtype cmd: cliff.command.Command
        :param result: return value of cmd
        :paramtype result: int
        :param err: exception or None
        :paramtype err: Exception
        """
        return

    def interact(self) -> None:
        # Defer importing .interactive as cmd2 is a slow import
        from .interactive import InteractiveApp

        if self.interactive_app_factory is None:
            interactive_app_factory = InteractiveApp
            self.interactive_app_factory = interactive_app_factory
        else:
            interactive_app_factory = self.interactive_app_factory
        self.interpreter = interactive_app_factory(
            self,
            self.command_manager,
            self.stdin,
            self.stdout,
        )
        self.interpreter.cmdloop()

    def get_fuzzy_matches(self, cmd: str) -> list[str]:
        """return fuzzy matches of unknown command"""

        sep = '_'
        if self.command_manager.convert_underscores:
            sep = ' '
        all_cmds = [k[0] for k in self.command_manager]
        dist = []
        for candidate in sorted(all_cmds):
            prefix = candidate.split(sep)[0]
            # Give prefix match a very good score
            if candidate.startswith(cmd):
                dist.append((0, candidate))
                continue
            # Levenshtein distance
            dist.append(
                (
                    utils.damerau_levenshtein(cmd, prefix, utils.COST) + 1,
                    candidate,
                )
            )

        matches = []
        match_distance = 0
        for distance, candidate in sorted(dist):
            if distance > match_distance:
                if match_distance:
                    # we copied all items with minimum distance, we are done
                    break
                # we copied all items with distance=0,
                # now we match all candidates at the minimum distance
                match_distance = distance
            matches.append(candidate)

        return matches

    def run_subcommand(self, argv: list[str]) -> int:
        try:
            subcommand = self.command_manager.find_command(argv)
        except ValueError as exc:
            # If there was no exact match, try to find a fuzzy match
            the_cmd = argv[0]
            fuzzy_matches = self.get_fuzzy_matches(the_cmd)
            if fuzzy_matches:
                article = 'a'
                if self.NAME[0] in 'aeiou':
                    article = 'an'
                self.stdout.write(
                    '{}: \'{}\' is not {} {} command. '
                    'See \'{} --help\'.\n'.format(
                        self.NAME,
                        ' '.join(argv),
                        article,
                        self.NAME,
                        self.NAME,
                    )
                )
                self.stdout.write('Did you mean one of these?\n')
                for match in fuzzy_matches:
                    self.stdout.write(f'  {match}\n')
            else:
                if self.options.debug:
                    raise
                else:
                    self.LOG.error(exc)
            return 2

        cmd_factory, cmd_name, sub_argv = subcommand
        kwargs = {}
        if 'cmd_name' in inspect.getfullargspec(cmd_factory.__init__).args:
            kwargs['cmd_name'] = cmd_name
        cmd = cmd_factory(self, self.options, **kwargs)
        result = 1
        err: BaseException | None = None
        try:
            self.prepare_to_run_command(cmd)
            full_name = (
                cmd_name
                if self.interactive_mode
                else ' '.join([self.NAME, cmd_name])
            )
            cmd_parser = cmd.get_parser(full_name)
            try:
                parsed_args = cmd_parser.parse_args(sub_argv)
            except SystemExit as ex:
                if self.interactive_mode:
                    # Defer importing cmd2 as it is a slow import
                    import cmd2

                    raise cmd2.exceptions.Cmd2ArgparseError from ex
                else:
                    raise ex
            result = cmd.run(parsed_args)
        except BrokenPipeError as err1:
            result = _SIGPIPE_EXIT
            err = err1
            raise
        except help.HelpExit:
            result = 0
        except Exception as err1:
            err = err1
            if self.options.debug:
                self.LOG.exception(err)
            else:
                self.LOG.error(err)
        except KeyboardInterrupt as err1:
            result = _SIGINT_EXIT
            err = err1
            raise
        finally:
            try:
                self.clean_up(cmd, result, err)
            except Exception as err2:
                if self.options.debug:
                    self.LOG.exception(err2)
                else:
                    self.LOG.error('Could not clean up: %s', err2)
            del err
        return result
